#include <stdlib.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>		/* IEEE Std 1003.1-2001 non-conformance. */
#include <sys/time.h>		/* IEEE Std 1003.1-2001 non-conformance. */
#include <sys/resource.h>

#include "config.h"
#include "imapfilter.h"


/* Secure memory information. */
typedef struct secmem {
	void *buf;		/* Allocated memory buffer. */
	size_t size;		/* Size of the buffer. */
	struct secmem *prev, *next;	/* Previous/next node of doubly linked
					 * list. */
} secmem_t;


extern unsigned int options;

static secmem_t *smem = NULL;	/* First node of secure memory linked list. */


void secmem_append(secmem_t * node);
secmem_t *secmem_find(void *ptr);
void secmem_remove(secmem_t * node);


/*
 * A malloc() that checks the results and dies in case of error.
 */
void *
xmalloc(size_t size)
{
	void *ptr;

	ptr = (void *)malloc(size);

	if (ptr == NULL)
		fatal(ERROR_MEMORY_ALLOCATION,
		    "allocating memory; %s\n", strerror(errno));

	return ptr;
}


/*
 * A realloc() that checks the results and dies in case of error.
 */
void *
xrealloc(void *ptr, size_t size)
{
	ptr = (void *)realloc(ptr, size);

	if (ptr == NULL)
		fatal(ERROR_MEMORY_ALLOCATION,
		    "allocating memory; %s\n", strerror(errno));

	return ptr;
}


/*
 * A free() that dies if fed with NULL pointer.
 */
void
xfree(void *ptr)
{
	if (ptr == NULL)
		fatal(ERROR_MEMORY_ALLOCATION,
		    "NULL pointer given as argument");
	free(ptr);
}


/*
 * A strdup() that checks the results and dies in case of error.
 */
char *
xstrdup(const char *s)
{
	char *cp;

	cp = strdup(s);

	if (cp == NULL)
		fatal(ERROR_MEMORY_ALLOCATION,
		    "allocating memory; %s\n", strerror(errno));

	return cp;
}


/*
 * Secure memory malloc().  Locks memory and keeps information about the
 * chunk that was allocated.
 */
void *
smalloc(size_t size)
{
	void *ptr;
	secmem_t *node;

	ptr = xmalloc(size);

	node = (secmem_t *) xmalloc(sizeof(secmem_t));

	node->buf = ptr;
	node->size = size;
	node->prev = node->next = NULL;

	secmem_append(node);

	return ptr;
}


/*
 * Secure memory realloc().  Resize memory by allocating a new memory chunk
 * and NULL fill old memory, in order to protect sensitive data.
 */
void *
srealloc(void *ptr, size_t size)
{
	void *p;
	secmem_t *node;

	if (!(node = (secmem_t *) secmem_find(ptr))) {
		ptr = xrealloc(ptr, size);
		return ptr;
	}
	p = smalloc(size);
	memcpy(p, node->buf, min(node->size, size));

	memset(node->buf, 0, node->size);
	secmem_remove(node);
	xfree(node->buf);
	xfree(node);

	return p;
}


/*
 * Secure memory free().  NULL fill memory before freeing it.
 */
void
sfree(void *ptr)
{
	secmem_t *node;

	if (!(node = (secmem_t *) secmem_find(ptr))) {
		xfree(ptr);
		return;
	}
	memset(node->buf, 0, node->size);
	secmem_remove(node);
	xfree(node->buf);
	xfree(node);
}


/*
 * Secure memory strdup().  Uses secure memory allocation.
 */
char *
sstrdup(const char *s)
{
	char *p;

	p = (char *)smalloc(strlen(s) + 1);
	xstrncpy(p, s, strlen(s));

	return p;
}


/*
 * Append information about the newly allocated memory buffer.
 */
void
secmem_append(secmem_t * node)
{
	secmem_t *pos;
	secmem_t **app;

	app = &smem;
	pos = smem;

	while (pos) {
		node->prev = pos;
		app = &(pos->next);
		pos = pos->next;
	}

	*app = node;
}


/*
 * Find the record of a memory buffer in the secure memory linked list.
 */
secmem_t *
secmem_find(void *ptr)
{
	secmem_t *pos;

	pos = smem;

	while (pos != NULL && pos->buf != ptr)
		pos = pos->next;

	return pos;
}


/*
 * Remove a record of a secure memory buffer.
 */
void
secmem_remove(secmem_t * node)
{
	if (node->prev != NULL)
		node->prev->next = node->next;
	if (node->next != NULL)
		node->next->prev = node->prev;
	if (smem == node)
		smem = node->next;

}


/*
 * Overwrite/clear all secure memory.
 */
void
secmem_clear(void)
{
	secmem_t *p, *t;

	for (p = smem; p != NULL; p = t) {
		t = p->next;
		sfree(p->buf);
	}
}


/*
 * Disable core file dumping.
 */
void
corefile_disable(void)
{
	struct rlimit rl;

	getrlimit(RLIMIT_CORE, &rl);

	rl.rlim_cur = rl.rlim_max = 0;
	setrlimit(RLIMIT_CORE, &rl);
}
